/**
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 *
 *  Authors:
 *  2009-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl> 
 *
 *  2007-2009
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *
 *  2003-2007
 *     Martijn Steenbakkers <martijn@nikhef.nl>
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "utils.h"

#include <lcmaps/lcmaps.h>


#ifndef LCMAPS_LIBDIR
    #define LCMAPS_LIBDIR ""
#endif
#ifndef LIBLCMAPS_SO
    #define LIBLCMAPS_SO "liblcmaps" LIBSUFF
#endif

static int set_liblcmaps_path(char ** liblcmaps_path);

static int set_liblcmaps_path(char ** liblcmaps_path)
{
    char * lcmaps_libdir       = NULL;
    char * lcmaps_moddir_sfx   = NULL;
    char * lcmaps_modules_dir  = NULL;
    struct stat st;

    /* First get the $LLGT_LCMAPS_LIBDIR environment variable as a base setting */
    lcmaps_libdir = getenv("LLGT_LCMAPS_LIBDIR");
    if (lcmaps_libdir && ((lcmaps_libdir)[0] != '\0')) {
        /* If it's non-empty, check it is an absolute and valid directory */
        if ((lcmaps_libdir)[0] != '/' || stat(lcmaps_libdir, &st)!=0 || !S_ISDIR(st.st_mode))  {
            llgt_logmsg(LOG_WARNING, "%s: Ignoring $LLGT_LCMAPS_LIBDIR as \"%s\" is not an absolute path to a valid directory\n", __func__, lcmaps_libdir);
            lcmaps_libdir = LCMAPS_LIBDIR;
        }
        /* Get the potentially different $LCMAPS_MODULE_SFX, based on the customized libdir */
        lcmaps_moddir_sfx = getenv("LLGT_LCMAPS_MODULEDIR_SFX");
        if (!lcmaps_moddir_sfx) {
            lcmaps_moddir_sfx = LCMAPS_MODULEDIR_SFX; /* Take the input from the LLGT configure */
        }

        /* Concatting LCMAPS_LIBDIR and LCMAPS_MODULEDIR_SFX to be exported as $LCMAPS_MODULES_DIR */
        lcmaps_modules_dir = malloc(AP_MAXPATH);
        if (!lcmaps_modules_dir) {
            llgt_logmsg(LOG_ERR, "%s: Could not allocate memory: %s\n", __func__, strerror(errno));
            return 1;
        } else {
            if (snprintf(lcmaps_modules_dir,AP_MAXPATH,"%s%s", lcmaps_libdir, lcmaps_moddir_sfx) >= AP_MAXPATH) {
                llgt_logmsg(LOG_WARNING, "Full modulespath '%s%s' would be too long, not setting LCMAPS_MODULES_DIR\n", lcmaps_libdir,lcmaps_moddir_sfx);
            } else {
                llgt_logmsg(LOG_DEBUG, "Setting LCMAPS_MODULES_DIR to '%s'\n", lcmaps_modules_dir);
                setenv("LCMAPS_MODULES_DIR", lcmaps_modules_dir, 1);
            }
            free(lcmaps_modules_dir);
        }

        /* Setting the full path to the liblcmaps.so object */
        *liblcmaps_path = malloc(AP_MAXPATH);
        if (snprintf(*liblcmaps_path, AP_MAXPATH, "%s/%s", lcmaps_libdir, LIBLCMAPS_SO) >= AP_MAXPATH) {
            llgt_logmsg(LOG_ERR, "Full path to %s \"%s/%s\" is too long\n", LIBLCMAPS_SO, lcmaps_libdir, LIBLCMAPS_SO);
            return 1;
        }
    } else {
        /* Just set the name of the shared object file, let the system figure it out */
        *liblcmaps_path = strdup(LCMAPS_LIBDIR LIBLCMAPS_SO);
    }
    llgt_logmsg(LOG_DEBUG, "LCMAPS library path : \"%s\"\n", *liblcmaps_path);
    return 0;
}


int llgt_run_lcmaps(gss_cred_id_t user_cred_handle, char * client_name, FILE * logfile, char ** username)
{
    /* USER MAPPING */
    void *  lcmaps_handle   = NULL;
    char *  error           = NULL;
    int     retval          = 0;
    char *  liblcmaps_path  = NULL;
    int (*LcmapsInit)(FILE *);
    int (*LcmapsTerm)();
    void (*LcmapsEnableVomsAttributesVerification)();
    void (*LcmapsDisableVomsAttributesVerification)();
#if ALLOW_EMPTY_CREDENTIALS
    int (*LcmapsRunAndReturnUsername)(char*, gss_cred_id_t, char*, char**, int, char**);
#else
    int (*LcmapsRunAndReturnUsername)(gss_cred_id_t, char*, char**, int, char**);
#endif /* ALLOW_EMPTY_CREDENTIALS */

    /* Get the path to the to be opened LCAS main library - Must be free'd */
    if (set_liblcmaps_path(&liblcmaps_path)) {
        llgt_logmsg(LOG_ERR, "Couldn't set the path to \"%s\"\n", LIBLCMAPS_SO);
        return 1;
    }

    /* I must have a path set */
    if (!liblcmaps_path) {
        llgt_logmsg(LOG_ERR, "Failed set a name or path to find liblcas.so\n");
        return 1;
    }

    lcmaps_handle = dlopen(liblcmaps_path, RTLD_LAZY|RTLD_GLOBAL);
    if (!lcmaps_handle) {
        llgt_logmsg(LOG_ERR, "Failed to dynamically load the library for LCMAPS: \"%s\"\n", LIBLCMAPS_SO);
        return 1;
    } else {
        /* Map the symbols */
        /* Init */
        LcmapsInit=dlsym(lcmaps_handle,"lcmaps_init");
        if ((error = dlerror()) != NULL) {
            llgt_logmsg(LOG_ERR, "LCMAPS module not compliant: Symbol \"%s\" not found: %s", "lcmaps_init", error);
            dlclose(lcmaps_handle);
            return 1;
        }
        /* Run */
        LcmapsRunAndReturnUsername=dlsym(lcmaps_handle,"lcmaps_run_and_return_username");
        if ((error = dlerror()) != NULL) {
            llgt_logmsg(LOG_ERR, "LCMAPS module not compliant: Symbol \"%s\" not found: %s", "lcmaps_run_and_return_username", error);
            dlclose(lcmaps_handle);
            return 1;
        }
        /* Term */
        LcmapsTerm=dlsym(lcmaps_handle,"lcmaps_term");
        if ((error = dlerror()) != NULL) {
            llgt_logmsg(LOG_ERR, "LCMAPS module not compliant: Symbol \"%s\" not found: %s", "lcmaps_term", error);
            dlclose(lcmaps_handle);
            return 1;
        }
        /* VOMS: Call to explicitly ENABLE the VOMS credential verification */
        LcmapsEnableVomsAttributesVerification=dlsym(lcmaps_handle,"lcmaps_enable_voms_attributes_verification");
        if ((error = dlerror()) != NULL) {
            llgt_logmsg(LOG_DEBUG, "LCMAPS module does not support explicit VOMS enable- or disablement. Symbol \"%s\" not found: %s", "lcmaps_enable_voms_attributes_verification", error);
        }
        /* VOMS: Call to explicitly DISABLE the VOMS credential verification */
        LcmapsDisableVomsAttributesVerification=dlsym(lcmaps_handle,"lcmaps_disable_voms_attributes_verification");
        if ((error = dlerror()) != NULL) {
            llgt_logmsg(LOG_DEBUG, "LCMAPS module does not support explicit VOMS enable- or disablement. Symbol \"%s\" not found: %s", "lcmaps_disable_voms_attributes_verification", error);
        }

        /* Initialize LCMAPS, do user mapping and terminate LCMAPS */
        retval=(*LcmapsInit)(logfile);
        if (retval)
        {
            llgt_logmsg(LOG_ERR, "LCMAPS initialization failure.");
            dlclose(lcmaps_handle);
            return 1;
        }

        /* VOMS: Check if the VOMS credential verification needs to be disabled
         * or explicitly enabled. Assuming there was a symbol in LCMAPS found
         * that is capable of performing the behavioural change. */
        if (LcmapsDisableVomsAttributesVerification && getenv("LLGT_VOMS_DISABLE_CREDENTIAL_CHECK")) {
            (*LcmapsDisableVomsAttributesVerification)();
        }
        if (LcmapsEnableVomsAttributesVerification && getenv("LLGT_VOMS_ENABLE_CREDENTIAL_CHECK")) {
            (*LcmapsEnableVomsAttributesVerification)();
        }

#if ALLOW_EMPTY_CREDENTIALS
        retval=(*LcmapsRunAndReturnUsername)(client_name,
                                             user_cred_handle,
                                             NULL,
                                             username,
                                             0,
                                             NULL
                                            );
#else
        retval=(*LcmapsRunAndReturnUsername)(user_cred_handle,
                                             NULL,
                                             username,
                                             0,
                                             NULL
                                            );
#endif
        if (retval) {
            llgt_logmsg(LOG_WARNING, "Warning: failed mapping. LCMAPS returned: %d\n", retval);
            if ((*LcmapsTerm)()) {
                llgt_logmsg(LOG_ERR, "LCMAPS Termination failure!");
                dlclose(lcmaps_handle);
                return 1;
            }
            dlclose(lcmaps_handle);
            return 1;
        }
        retval=(*LcmapsTerm)();
        if (retval) {
            llgt_logmsg(LOG_ERR, "LCMAPS Termination failure!");
            dlclose(lcmaps_handle);
            return 1;
        }
        dlclose(lcmaps_handle);
    }
    /* END USER MAPPING */
    return 0;
}
